"""
Package version and distribution
"""
import pkg_resources as pkr

__pkg__ = 'avotes-parser-cli'

try:
    __distribution__ = pkr.get_distribution(__pkg__)
    __version__ = __distribution__.version
except pkr.DistributionNotFound:
    __distribution__ = pkr.Distribution()
    __version__ = '0.0.0'



"""
Parsing and decoding aragon votes.

Prepare human-readable representation of the last N votes
for a aragon application with a specific address in a target net.
"""
import argparse
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial

from brownie.utils import color

from avotes_parser.core.ABI import get_cached_etherscan_api
from avotes_parser.core.ABI.utilities.etherscan import (
    DEFAULT_NET, NET_URL_MAP
)
from .utilities import (
    logging_settings, read_key, get_aragon_voting,
    parse_voting, set_infura_project_id
)

DEFAULT_ARAGON_ADDRESS = '0x2e59A20f205bB85a89C53f1936454680651E618e'
DEFAULT_N = 10
DEFAULT_WORKERS_NUM = DEFAULT_N


def parse_args() -> argparse.Namespace:
    """Parse CLI arguments."""
    parser = argparse.ArgumentParser(
        add_help=True,
        description=__doc__,
        prog='avotes-parser',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument('--apitoken',
                        required=True,
                        type=str,
                        help='Etherscan API key as string or '
                             'a path to txt file with it.')
    parser.add_argument('--infura',
                        type=str,
                        required=True,
                        help='Infura project ID.')

    parser.add_argument('-n',
                        type=int,
                        default=DEFAULT_N,
                        help='Parse last N votes.')
    parser.add_argument('--aragon-voting-address',
                        type=str,
                        default=DEFAULT_ARAGON_ADDRESS,
                        help='Address of aragon voting contract.')
    parser.add_argument('--net',
                        type=str,
                        default=DEFAULT_NET,
                        help='Net name is case-insensitive.',
                        choices=NET_URL_MAP.keys())
    parser.add_argument('--retries',
                        type=int,
                        default=5,
                        help='Number of retries of calling Etherscan API.')
    parser.add_argument('--num-workers',
                        default=DEFAULT_WORKERS_NUM,
                        type=int,
                        help='Number of asynchronous parsing tasks.')
    parser.add_argument('--debug',
                        action='store_true',
                        help='Show debug messages')

    return parser.parse_args()


def main():
    """Parse and decode aragon votes."""
    args = parse_args()

    logging_settings(args.debug)
    token = read_key(args.apitoken)
    set_infura_project_id(args.infura)
    logging.debug(f'API key: {token}')

    aragon_voting = get_aragon_voting(
        args.net, args.aragon_voting_address,
        token, args.retries
    )

    last_vote_number = aragon_voting.votesLength()

    abi_storage = get_cached_etherscan_api(
        token, args.net
    )

    _parse_voting = partial(parse_voting, aragon_voting, abi_storage)
    number_offset = last_vote_number - args.n
    with ThreadPoolExecutor(args.num_workers) as executor:
        parsed_votes = {
            executor.submit(
                _parse_voting,
                vote_number + number_offset
            ): vote_number
            for vote_number in range(
                0, args.n
            )
        }
        results = [None] * len(parsed_votes)
        for future in as_completed(parsed_votes):
            number = parsed_votes[future]
            results[number] = future.result()

    for num, parsed_vote in enumerate(results):
        number = num + number_offset
        total = len(parsed_vote)
        print(f'Voting #{number}.')
        for ind, call in enumerate(parsed_vote):
            print(f'Point {ind + 1}/{total}.')
            print(color.highlight(repr(call)))
        print('------------------------------------------------\n\n')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['avotes_parser', 'avotes_parser.cli']

package_data = \
{'': ['*']}

install_requires = \
['avotes-parser-core>=0.5.2,<0.6.0',
 'eth-brownie>=1.19.1,<1.20.0',
 'pysha3>=1.0.2,<1.1.0',
 'requests>=2.28.1,<2.29.0',
 'web3>=5.30.0,<5.31.0']

entry_points = \
{'console_scripts': ['avotes-parser = avotes_parser.cli.__main__:main']}

setup_kwargs = {
    'name': 'avotes-parser-cli',
    'version': '0.5.2',
    'description': 'Aragon votings parser CLI',
    'long_description': None,
    'author': 'Dmitri Ivakhnenko',
    'author_email': 'dmit.ivh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
Utilities and helpers of CLI tool.
"""
import logging
import os
from functools import lru_cache
from mimetypes import guess_type, types_map
from typing import Union, List

import brownie

from avotes_parser.core import (
    Call, decode_function_call, parse_script
)
from avotes_parser.core.ABI.storage import CachedStorage
from avotes_parser.core.ABI.utilities.etherscan import (
    get_implementation_address, get_abi
)
from avotes_parser.core.ABI.utilities.exceptions import (
    ABIEtherscanNetworkError, ABIEtherscanStatusCode
)
from avotes_parser.core.parsing import ParseStructureError


def set_infura_project_id(project_id: str) -> None:
    """Set WEB3_INFURA_PROJECT_ID environment variable."""
    os.environ['WEB3_INFURA_PROJECT_ID'] = project_id


def read_key(path_to: str) -> str:
    """Read Etherscan API key."""
    m_type, _ = guess_type(path_to)
    if m_type == types_map['.txt']:
        with open(path_to, 'r') as api_token_file:
            return api_token_file.read().strip()

    else:
        return path_to


def logging_settings(debug: bool) -> None:
    """Prepare logger."""
    if debug:
        level = logging.DEBUG

    else:
        level = logging.INFO

    logging.basicConfig(
        format='%(levelname)s:%(message)s', level=level
    )


@lru_cache
def get_aragon_voting(
        net: str, address: str,
        etherscan_api_key: str, retries: int,
):
    """Get aragon voting contract as object."""
    if not brownie.network.is_connected():
        brownie.network.connect(net)

    abi = get_abi(etherscan_api_key, address, net, retries)
    impl_address = get_implementation_address(
        address, abi, net
    )
    impl_abi = get_abi(etherscan_api_key, impl_address, net, retries)
    return brownie.Contract.from_abi(
        'AragonVoting', address, impl_abi
    )


def decode_evm_script(
        script: str,
        abi_storage: CachedStorage
) -> List[Union[Call, str]]:
    """
    Parse and decode EVM script.
    """
    try:
        parsed = parse_script(script)
    except ParseStructureError as err:
        return [repr(err)]

    calls = []
    for call in parsed.calls:
        try:
            decoded_call = decode_function_call(
                call.address, call.method_id,
                call.encoded_call_data, abi_storage
            )
            if decoded_call is None:
                calls.append(call)
            else:
                calls.append(decoded_call)
                for inp in decoded_call.inputs:
                    if inp.type == 'bytes' and inp.name == '_evmScript':
                        inp.value = decode_evm_script(inp.value, abi_storage)
                        break

        except (ABIEtherscanNetworkError, ABIEtherscanStatusCode) as err:
            calls.append(f'Network layer error: {repr(err)}')

    return calls


def parse_voting(
        aragon_voting, abi_storage: CachedStorage,
        vote_number: int
) -> List[Union[Call, str]]:
    """Decode aragon voting with specific number."""
    script_code = str(aragon_voting.getVote(vote_number)[-1])
    return decode_evm_script(script_code, abi_storage)
